#include "sp_multi_content.h"

/*******************************************************
* This plugin checks for a content match across packets
* if a content match is made at the beginning or end
* of a packet, the cache is called and the two packets
* reassymbled to see if the content matches.
********************************************************/

/*******************************************************
* Push a function on the detection engine list
*******************************************************/
void MultiContentInit(char* data, OptTreeNode* otn, int protocol){
	    /* allocate the data structure and attach it to the
       rule's data struct list */
    otn->ds_list[PLUGIN_MULTI_CONTENT] = (MultiContentData *) calloc(sizeof(MultiContentData), sizeof(char));
	
    /* this is where the keyword arguments are processed and placed into the 
       rule option's data structure */
    ParseMultiContent(data, otn);
}

/*******************************************************
* Push a function on the detection engine list
*******************************************************/
void NoCaseMultiContentInit(char* data, OptTreeNode* otn, int protocol){
	    /* allocate the data structure and attach it to the
       rule's data struct list */
    otn->ds_list[PLUGIN_MULTI_CONTENT] = (MultiContentData *) calloc(sizeof(MultiContentData), sizeof(char));
	((MultiContentData *)otn->ds_list[PLUGIN_MULTI_CONTENT])->flags = MULTI_CONTENT_FLAG_NOCASE;

    /* this is where the keyword arguments are processed and placed into the 
       rule option's data structure */
    ParseMultiContent(data, otn);
	
}

/*******************************************************
* Register the plugin with the engine
*******************************************************/
void SetupMultiContent(){
    RegisterPlugin("multi", MultiContentInit);
	RegisterPlugin("multi-nocase", NoCaseMultiContentInit);

#ifdef DEBUG
    printf("Plugin: MultiContent Initialized\n");
#endif
}

/*******************************************************
*
*******************************************************/
void ParseMultiContent(char* data, OptTreeNode* otn){
	
	MultiContentData*	ds_ptr;
	char*				this;
	char*				raw;

	/*alloc the struct*/
	ds_ptr = (MultiContentData*)otn->ds_list[PLUGIN_MULTI_CONTENT];
	
	/*kill the whitespace*/
	while (isspace((int)*data)) data++;
	
#ifdef DEBUG
	printf("*************************************\n");
	printf("*Parsing Multi:%s\n", data);
	printf("*************************************\n");
#endif
	
	ds_ptr->data=(char*)calloc(strlen(data), sizeof(char));
	this=data;
	raw=ds_ptr->data;
	while (*this){
		switch (*this){
		case '"':
			/*ignore this*/
			break;
		case '\\':
			/*copy the next char as a literal*/
			this++;
			if (ds_ptr->flags & MULTI_CONTENT_FLAG_NOCASE){
				*raw=toupper(*this);
			}else{
				*raw=*this;
			}
			raw++;
			ds_ptr->len++;
			break;	
		default:
			if (ds_ptr->flags & MULTI_CONTENT_FLAG_NOCASE){
				*raw=toupper(*this);
			}else{
				*raw=*this;
			}
			raw++;
			ds_ptr->len++;
		}
		this++;
	}
	
#ifdef DEBUG
	printf("Result of raw:%s\n", ds_ptr->data);
	printf("Result len is %u\n", ds_ptr->len);
#endif	
	
	AddOptFuncToList(CheckMultiContent, otn);
}

/*******************************************************
* return TRUE if the two chars are the same after
* case conversions etc.
*******************************************************/
int CharEqual(char c1, char c2, int flags){
	if (flags & MULTI_CONTENT_FLAG_NOCASE){
		return (toupper(c1)==c2);
	}else{
		return (c1==c2);
	}
}

/*******************************************************
* Check for a full or partial signature match
*******************************************************/
int CheckMultiContent(Packet* p, struct _OptTreeNode* otn, OptFpList* fp_list){
	register char*	test;
	register char*	sig;
	int				testlen;
	int				siglen;
	int				flags;
	int				i,j;
	PacketHolder*	ph;
	int				ph_len;
	
#ifdef DEBUG
	printf("Checking MultiContent\n");
#endif

	/*if there's nothing to do, bail*/
	if (!p->dsize) return 0;
	
	testlen = p->dsize;
	siglen = ((MultiContentData*)otn->ds_list[PLUGIN_MULTI_CONTENT])->len;
	test=p->data;
	sig=((MultiContentData*)otn->ds_list[PLUGIN_MULTI_CONTENT])->data;
	flags=((MultiContentData*)otn->ds_list[PLUGIN_MULTI_CONTENT])->flags;
	*(test+testlen)=0x00;
	*(sig+siglen)=0x00;
	
	/*walk backwards until we find the last char*/
	for (i=testlen; i>=0 ; i--){
		if (CharEqual(*(test+i),*(sig+siglen-1), flags)){  
			/*matched last letter. Now check the rest*/
			for (j=0; j<siglen; j++){
				if (!CharEqual(*(test+i-j),*(sig+siglen-j-1), flags)) break;
				if (j==siglen-1){
#ifdef DEBUG
					printf("Hard Match\n");
#endif				
					/*call the next function in the list*/
					return fp_list->next->OptTestFunc(p, otn, fp_list->next);
				}
				if (i==j){
					/*we only reassymble tcp packets*/
					if (!p->tcph) return 0;
#ifdef DEBUG
					printf("Partial Match at front\n");
#endif
					/*get the last packet and compare*/
					/*return 0 for now*/
					ph=GetCachePacket(p->state, ntohl(p->tcph->th_seq));
					if (!ph) return 0;
					
					/*see of the signature matches across packets*/
					ph_len=siglen-j-1;
					test=ph->payload + (PACKET_SIZE-ph_len);
					for (i=0;i<ph_len;i++){
						if (!CharEqual(*(test+i),*(sig+i), flags))
							return 0;
					}
#ifdef DEBUG					
					printf("Matched!\n");
#endif					
					return fp_list->next->OptTestFunc(p, otn, fp_list->next);					
				}
			}
		}
	}
#ifdef DEBUG	
	printf("Not Found\n");
#endif	
	/*check for partial match at end*/
	/*if found, cache this packet*/
	/*we can only resassymble tcp packets*/
	if (!p->tcph) return 0;
	
	for (i=0; i< (siglen-1); i++){
		if (i>testlen) break;
		if (CharEqual(*(test+testlen-i), *sig, flags)){
			for (j=testlen-i; j<testlen; j++){
				if (!CharEqual(*(test+j), *(sig+j-(testlen-i)), flags)) break;
			}
			if (j==testlen){
#ifdef DEBUG			
				printf("Partial signature match at end\n");
#endif				
				/*cache this packet for later*/
				CachePacket(p->state, p);
				break;
			}
		}
	}

	return 0;
}
